#include "hc16.hpp"

static bool flow;               // does the current instruction pass
// execution to the next instruction?

//------------------------------------------------------------------------
// Handle an operand with an immediate value:
//      - mark it with FF_IMMD flag
//      - for bit logical instructions specify the operand type as a number
//        because such an operand is likely a plain number rather than
//        an offset or of another type.

static void doImmdValue(void)
{
	doImmd(cmd.ea);
	uFlag = getFlags(cmd.ea);             // refresh uFlag with new flags
}

//----------------------------------------------------------------------
// Calculate the target data address
ea_t map_addr(ulong off, int opnum, bool isdata)
{
	if ( isdata )
	{
		//    if ( isOff(uFlag, opnum) ) 
		//		return get_offbase(cmd.ea, opnum) >> 4;
		return intmem + off;
	}
	return toEA(codeSeg(off, opnum), off);
}

//----------------------------------------------------------------------
// Handle an operand. What this function usually does:
//      - creates cross-references from the operand
//        (the kernel deletes all xrefs before calling emu())
//      - creates permanent comments
//      - if possible, specifies the operand type (for example, it may
//        create stack variables)
//      - anything else you might need to emulate or trace

static void handle_operand(op_t &x,int loading)
{
	switch ( x.type )
	{
	case o_phrase:              // no special hanlding for these types
	case o_reg:
		break;

	case o_imm:                         // an immediate number as an operand
		doImmdValue();                    // handle immediate number

		// if the value was converted to an offset, then create a data xref:
		if ( isOff(uFlag, x.n) ) 
			ua_add_off_drefs(x, dr_O);

		break;

	case o_displ:
		doImmdValue();                    // handle immediate number

		// if the value was converted to an offset, then create a data xref:
		if ( isOff(uFlag, x.n) ) 
			ua_add_off_drefs(x, dr_O);
		break;

	case o_mem:                         // an ordinary memory data reference
		{
			ea_t dea = map_addr(x.addr, x.n, true);
			ua_dodata(dea, x.dtyp);
			if ( !loading ) 
				doVar(dea);     // write access
			ua_add_dref(x.offb, dea, loading ? dr_R : dr_W);
		}
		break;

	case o_far:
	case o_near:                        // a code reference
		{
			ea_t ea = map_addr(x.addr, x.n, false);
			int iscall = InstrIsSet(cmd.itype, CF_CALL);
			ua_add_cref(x.offb, ea, iscall ? fl_CN : fl_JN);
			if ( flow && iscall )
			{
				func_t *pfn = get_func(ea);
				if ( pfn != NULL && (pfn->flags & FUNC_NORET) ) 
					flow = false;
			}
		}
		break;

	default:
		warning("%a (%s): bad optype", cmd.ea, Instructions[cmd.itype].name);
		break;
	}
}

//----------------------------------------------------------------------
// Emulate an instruction
// This function should:
//      - create all xrefs from the instruction
//      - perform any additional analysis of the instruction/program
//        and convert the instruction operands, create comments, etc.
//      - create stack variables
//      - analyse the delayed branches and similar constructs
// The kernel calls ana() before calling emu(), so you may be sure that
// the 'cmd' structure contains a valid and up-to-date information.
// You are not allowed to modify the 'cmd' structure.
// Upon entering this function, the 'uFlag' variable contains the flags of
// cmd.ea. If you change the characteristics of the current instruction, you
// are required to refresh 'uFlag'.
// Usually the kernel calls emu() with consecutive addresses in cmd.ea but
// you can't rely on this - for example, if the user asks to analyse an
// instruction at arbirary address, his request will be handled immediately,
// thus breaking the normal sequence of emulation.
// If you need to analyse the surroundings of the current instruction, you
// are allowed to save the contents of the 'cmd' structure and call ana().
// For example, this is a very common pattern:
//  {
//    insn_t saved = cmd;
//    if ( decode_prev_insn(cmd.ea) != BADADDR )
//    {
//      ....
//    }
//    cmd = saved;
//  }
//
// This sample emu() function is a very simple emulation engine.

int emu(void)
{
	int Feature = Instructions[cmd.itype].feature;
	flow = ((Feature & CF_STOP) == 0);

	if ( Feature & CF_USE1 ) 
		handle_operand(cmd.Op1, 1);
	if ( Feature & CF_USE2 ) 
		handle_operand(cmd.Op2, 1);
	if ( Feature & CF_USE3 ) 
		handle_operand(cmd.Op3, 1);
	if ( Feature & CF_JUMP ) 
		QueueMark(Q_jumps,cmd.ea);
	if ( Feature & CF_CALL ) 
		QueueMark(Q_jumps,cmd.ea);
	if ( Feature & CF_CHG1 ) 
		handle_operand(cmd.Op1, 0);
	if ( Feature & CF_CHG2 ) 
		handle_operand(cmd.Op2, 0);
	if ( Feature & CF_CHG3 ) 
		handle_operand(cmd.Op3, 0);

	// if the execution flow is not stopped here, then create
	// a xref to the next instruction.
	// Thus we plan to analyse the next instruction.

	if ( flow ) 
		ua_add_cref(0,cmd.ea+cmd.size,fl_F);

	// look for instructions that modify segment registers
	// todo: this isn't working properly yet
	/*
	sel_t tseg;
	switch(cmd.itype)
	{
	// things dealing with IX

	case HC16_ABX:	// Add B to IX 
	case HC16_ADX:	// Add D to IX 
	case HC16_AEX:	// Add E to IX 
	case HC16_XGDX:	// Exchange D with IX 
	case HC16_XGEX:	// Exchange E with IX 
	splitSRarea1(get_item_end(cmd.ea),rIX,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_AIX:	// Add Immediate Value to IX
	tseg = getSR(cmd.ea,rIX);
	if (tseg != BADSEL)
	splitSRarea1(get_item_end(cmd.ea),rIX,tseg+cmd.Op1.value,SR_auto);	// we don't know the value!
	break;
	case HC16_LDX:	// Load IX

	splitSRarea1(get_item_end(cmd.ea),rIX,tseg+cmd.Op1.value,SR_auto);	// we don't know the value!

	case HC16_TSX:	// Transfer SP to X 
	case HC16_TYX:	// Transfer Y to X 
	case HC16_TZX:	// Transfer Z to X 

	// things dealing with IY

	case HC16_ABY:	// Add B to IY 
	case HC16_ADY:	// Add D to IY 
	case HC16_AEY:	// Add E to IY 
	case HC16_XGDY:	// Exchange D with IY 
	case HC16_XGEY:	// Exchange E with IY 
	splitSRarea1(get_item_end(cmd.ea),rIY,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_AIY:	// Add Immediate Value to IY
	case HC16_LDY:	// Load IY 
	case HC16_TXY:	// Transfer X to Y 
	case HC16_TSY:	// Transfer SP to Y 
	case HC16_TZY:	// Transfer Z to Y

	// things dealing with IZ

	case HC16_ABZ:	// Add B to IZ
	case HC16_ADZ:	// Add D to IZ 		
	case HC16_AEZ:	// Add E to IZ 
	case HC16_XGDZ:	// Exchange D with IZ 
	case HC16_XGEZ:	// Exchange E with IZ 
	splitSRarea1(get_item_end(cmd.ea),rIZ,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_AIZ:	// Add Immediate Value to IZ
	case HC16_LDZ:	// Load IZ 
	case HC16_TSZ:	// Transfer SP to Z 
	case HC16_TXZ:	// Transfer X to Z 
	case HC16_TYZ:	// Transfer Y to Z 

	case HC16_TBEK:	// Transfer B to EK 
	splitSRarea1(get_item_end(cmd.ea),rEK,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_TBSK:	// Transfer B to SK 
	splitSRarea1(get_item_end(cmd.ea),rSK,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_TBXK:	// Transfer B to XK 
	splitSRarea1(get_item_end(cmd.ea),rXK,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_TBYK:	// Transfer B to YK 
	splitSRarea1(get_item_end(cmd.ea),rYK,BADSEL,SR_auto);	// we don't know the value!
	break;

	case HC16_TBZK:	// Transfer B to ZK 
	splitSRarea1(get_item_end(cmd.ea),rZK,BADSEL,SR_auto);	// we don't know the value!
	break;
	}
	*/
	return 1;    // actually the return value is unimportant, but let's it be so
}
