#include "hc16.hpp"
#include <fpro.h>
#include <diskio.hpp>

//----------------------------------------------------------------------
static void vlog(const char *format, va_list va)
{
	static FILE *fp = NULL;
	if ( fp == NULL ) fp = fopenWT("debug_log");
	qvfprintf(fp, format, va);
	qflush(fp);
}
//----------------------------------------------------------------------
inline void log(const char *format, ...)
{
	va_list va;
	va_start(va, format);
	vlog(format, va);
	va_end(va);
}

#define AT   COLSTR("@", SCOLOR_SYMBOL)
#define PLUS COLSTR("+", SCOLOR_SYMBOL)

static const char *phrases[] =
{
	AT COLSTR("R0", SCOLOR_REG),
		AT COLSTR("R1", SCOLOR_REG),
		AT COLSTR("DPTR", SCOLOR_REG),
		AT COLSTR("A", SCOLOR_REG) PLUS COLSTR("DPTR", SCOLOR_REG),
		AT COLSTR("A", SCOLOR_REG) PLUS COLSTR("PC", SCOLOR_REG)
};

//----------------------------------------------------------------------
inline void OutReg(int rgnum)
{
	out_register(ph.regNames[rgnum]);
}

//----------------------------------------------------------------------
// generate the text representation of an operand

bool outop(op_t &x)
{
	uval_t v;
	const char *ptr;
	int mask;
	switch ( x.type )
	{
	case o_reg:
		OutReg(x.reg);
		break;

	case o_phrase:
		out_symbol('@');
		OutReg(x.phrase);
		out_symbol('+');
		out_symbol('+');
		OutValue(x, OOF_ADDR | OOFS_IFSIGN | OOFW_8);
		break;
		/* 
		case o_displ:
		out_symbol('@');
		OutReg(x.reg);
		out_symbol('+');
		mask = OOF_ADDR | OOFS_IFSIGN;
		switch (x.dtyp)
		{
		case dt_byte:
		mask |= OOFW_8;
		break;
		case dt_word:
		mask |= OOFW_16;
		break;
		case dt_dword:
		mask |= OOFW_32;
		break;
		}
		OutValue(x,mask);
		break;
		*/
	case o_imm:
		out_symbol('#');
		mask = OOFS_IFSIGN | OOFW_IMM;
		switch (x.dtyp)
		{
		case dt_byte:
			mask |= OOFW_8;
			break;
		case dt_word:
			mask |= OOFW_16;
			break;
		case dt_dword:
			mask |= OOFW_32;
			break;
		}
		OutValue(x,mask);
		break;

	case o_mem:
	case o_far:
	case o_near:
	case o_displ:
		v = map_addr(x.addr, x.n, x.type==o_mem);
		ptr = get_name_expr(cmd.ea+x.offb, x.n, v, x.addr);
		if( ptr == NULL )
		{
			OutValue(x, OOF_ADDR | OOF_NUMBER | OOFS_NOSIGN | OOFW_32);
			QueueMark(Q_noName, cmd.ea);
			break;
		}
		OutLine(ptr);
		break;

	case o_void:
		return false;

	default:
		warning("out: %a: bad optype",cmd.ea,x.type);
		break;
	}
	return true;
}

//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure

void out(void)
{
	char buf[MAXSTR];

	init_output_buffer(buf, sizeof(buf)); // setup the output pointer
	OutMnem();                            // output instruction mnemonics

	out_one_operand(0);                   // output the first operand

	for (int i = 1; i < UA_MAXOP; i++)
		if ( cmd.Operands[i].type != o_void)
		{
			out_symbol(',');
			OutChar(' ');
			out_one_operand(i);                 // output the second operand
		}

		// output a character representation of the immediate values
		// embedded in the instruction as comments
		for (i = 0; i < UA_MAXOP; i++)
			if ( isVoid(cmd.ea,uFlag,i) ) 
				OutImmChar(cmd.Operands[i]);

		term_output_buffer();                 // terminate the output string
		gl_comm = 1;                          // ask to attach a possible user-
		// defined comment to it
		MakeLine(buf);                        // pass the generated line to the
		// kernel
}

//--------------------------------------------------------------------------
// generate start of the disassembly

void header(void)
{
	gen_cmt_line("Processor:        %s [%s]", device[0] ? device : inf.procName, deviceparams);
	gen_cmt_line("Processor:        %s", inf.procName);
	gen_cmt_line("Target assembler: %s", ash.name);
	if ( ash.header != NULL )
		for ( const char **ptr=ash.header; *ptr != NULL; ptr++ ) MakeLine(*ptr,0);
}

//--------------------------------------------------------------------------
// generate start of a segment

void segstart(ea_t ea)
{
	char buf[MAXSTR];
	segment_t *Sarea = getseg(ea);
	if ( ash.uflag & UAS_SECT ) {
		if ( Sarea->type == SEG_IMEM ) 
			MakeLine(".RSECT");
		else {
			qsnprintf(buf, sizeof(buf), COLSTR("%s: .section", SCOLOR_ASMDIR),
				get_segm_name(Sarea));
			MakeLine(buf,0);
		}
	} 
	else
	{
		if(ash.uflag & UAS_NOSEG)
			qsnprintf(buf, sizeof(buf), COLSTR("%s.segment %s", SCOLOR_AUTOCMT),ash.cmnt, get_segm_name(Sarea));
		else 
			qsnprintf(buf, sizeof(buf), COLSTR("segment %s",SCOLOR_ASMDIR),get_segm_name(Sarea));
		MakeLine(buf);
		if(ash.uflag & UAS_SELSG) 
			MakeLine(get_segm_name(Sarea));
		if(ash.uflag & UAS_CDSEG)
			MakeLine((Sarea->type == SEG_IMEM) ?
			COLSTR("DSEG", SCOLOR_ASMDIR) : COLSTR("CSEG", SCOLOR_ASMDIR));
	}
	if ( inf.s_org )
	{
		adiff_t org = ea - get_segm_base(Sarea);
		if( org != 0 ) 
			gen_cmt_line("%s %s", ash.origin, btoa32(org));
	}
}

//--------------------------------------------------------------------------
// generate end of the disassembly

void footer(void)
{
	char buf[MAXSTR];
	if ( ash.end != NULL)
	{
		MakeNull();
		char *const end = buf + sizeof(buf);
		char *ptr = tag_addstr(buf, end, COLOR_ASMDIR, ash.end);
		char name[MAXSTR];
		if ( get_colored_name(BADADDR, inf.beginEA, name, sizeof(name)) != NULL )
		{
			APPCHAR(ptr, end, ' ');
			if( ash.uflag & UAS_NOENS )
				APPEND(ptr, end, ash.cmnt);
			APPEND(ptr, end, name);
		}
		MakeLine(buf);
	}
	else
	{
		gen_cmt_line("end of file");
	}
}

//--------------------------------------------------------------------------
// generate a data representation
// usually all the job is handled by the kernel's standard procedure,
// intel_data()

void hc16_data(ea_t ea)
{
	gl_name = 1;
	intel_data(ea);
}
